// MC.CPP
//

#include "mc.h"

#include <oplerr.h>
#include <libc/stdio.h>
#include <f32file.h>


// void _decrypt(TInt32 *data, TInt32 size, TInt32 key);
TInt32 is_header_legal(MCEXTHEADER *header,TInt32 filesize);
TInt Lookup(MCEXTTABLE *t, TPtrC8 name, TInt32 *value);
TInt LookupN(MCEXTTABLE *t, TInt32 idx, TInt32 *value);
#ifdef SUPPORT_IMPORT
TInt link_symbol(MCEXTTABLE *import, MCEXTTABLE *symbol, MCEXTHEADER *header, TInt b);
TInt link(TInt32 *extlist, MCEXTTABLE *import, MCEXTHEADER *header, TInt b);
#endif
TInt _strcmp(const TText8 *t1, const TText8 *t2, const TInt len2);
MCEXTTABLE *skip_symbol(MCEXTTABLE *symbol);


///////////////////////////////////////////////////////////////////////////
// COpxRoot class : derives from COpxBase
//

// The language extension procedures provided by this OPX
//

void COpxRoot::MCOplAddr() const
	{ // return logical address
	TInt32 value = iOplAPI.PopInt32();
	value = (TInt32)iOplAPI.OffsetToAddrL((TInt)value, 0);
	iOplAPI.Push(value);
	}


void COpxRoot::MCPokeL() const
	{ // return 0
	TInt32 value = iOplAPI.PopInt32();
	TInt32 *addr = (TInt32 *)iOplAPI.PopInt32();
	if (addr != 0)  *addr = value;
	iOplAPI.Push((TInt32)0);
	}


void COpxRoot::MCPokeW() const
	{ // return 0
	TInt32 value = iOplAPI.PopInt32();
	TInt16 *addr = (TInt16 *)iOplAPI.PopInt32();
	if (addr != 0)  *addr = (value & 0xffff);
	iOplAPI.Push((TInt32)0);
	}


void COpxRoot::MCPokeB() const
	{ // return 0
	TInt32 value = iOplAPI.PopInt32();
	TInt8 *addr = (TInt8 *)iOplAPI.PopInt32();
	if (addr != 0)  *addr = (value &255);
	iOplAPI.Push((TInt32)0);
	}


void COpxRoot::MCPokeS() const
	{ // return 0
	TPtrC8 value = iOplAPI.PopString8();
	TText8 *addr = (TText8 *)iOplAPI.PopInt32();
	iOplAPI.Push((TInt32)0);
	}


void COpxRoot::MCPeekL() const
	{ // return value
	TInt32 *addr = (TInt32 *)iOplAPI.PopInt32();
	iOplAPI.Push(*addr);
	}


void COpxRoot::MCPeekW() const
	{ // return value
	TInt16 *addr = (TInt16 *)iOplAPI.PopInt32();
	iOplAPI.Push((TInt32)*addr);
	}


void COpxRoot::MCPeekB() const
	{ // return value
	TInt8 *addr = (TInt8 *)iOplAPI.PopInt32();
	iOplAPI.Push((TInt32)*addr);
	}


void COpxRoot::MCCallAddr() const
	{
	// call address, return value
	union {
	  TInt32 (* code)(TInt32, TInt32, TInt32, TInt32);
	  TInt32 value;
	} code;

	TInt32 r3 = iOplAPI.PopInt32();
	TInt32 r2 = iOplAPI.PopInt32();
	TInt32 r1 = iOplAPI.PopInt32();
	TInt32 r0 = iOplAPI.PopInt32();
	code.value = iOplAPI.PopInt32();

	iOplAPI.Push(code.code(r0, r1, r2, r3));
	}


void COpxRoot::MCCall() const
	{
	// call numbered function, return -1 or value
	union {
	  TInt32 (* code)(TInt32, TInt32, TInt32, TInt32, TInt8 *);
	  TInt32 value;
	} code;

	TInt32 r3 = iOplAPI.PopInt32();
	TInt32 r2 = iOplAPI.PopInt32();
	TInt32 r1 = iOplAPI.PopInt32();
	TInt32 r0 = iOplAPI.PopInt32();
	TInt32 idx = iOplAPI.PopInt32();
	MCEXT *handle = (MCEXT *)(iOplAPI.PopInt32() - sizeof(MCEXT));

	MCEXTHEADER *header = handle->header;

	if ((!handle->initialised) || (idx < -2)) {
		iOplAPI.Push((TInt32)-1);
		return;
	}

        // find address to call
        code.value = -1;

        if (idx < 64)         // use the buffered address
          code.value = handle->functions[idx];
	else if (idx == -1) {
	     if (header->init.offset)   code.value = header->init.offset;
	} else if (idx == -2) {
	     if (header->kill.offset)   code.value = header->kill.offset;
        } else {
          // lookup address
  	  if (LookupN((MCEXTTABLE *)header->functions, idx, &code.value))
                  code.value = -1;
        }

        TInt32 rval = -1;
        if (code.value != -1)
	     rval = code.code(r0, r1, r2, r3, handle->workspace);

        iOplAPI.Push(rval);
	}


void COpxRoot::MCCallF() const
	{
	// call named function, return -1 or value
	union {
	  TInt32 (* code)(TInt32, TInt32, TInt32, TInt32, TInt8 *);
	  TInt32 value;
	} code;

	TInt32 r3 = iOplAPI.PopInt32();
	TInt32 r2 = iOplAPI.PopInt32();
	TInt32 r1 = iOplAPI.PopInt32();
	TInt32 r0 = iOplAPI.PopInt32();
	TPtrC8 fname = iOplAPI.PopString8();
	MCEXT *handle = (MCEXT *)(iOplAPI.PopInt32() - sizeof(MCEXT));

	if (!handle->initialised) {
		iOplAPI.Push((TInt32)-1);
		return;
	}

	MCEXTHEADER *header = handle->header;

	if (Lookup((MCEXTTABLE *)header->functions, fname, &code.value)) {
		iOplAPI.Push((TInt32)-1);
		return;
	}

	TInt32 rval = code.code(r0, r1, r2, r3, handle->workspace);
	iOplAPI.Push(rval);
	}


void COpxRoot::MCLookup() const
	{
	  // return -1 or index of named function
	TPtrC8 funcname = iOplAPI.PopString8();
	MCEXT *handle = (MCEXT *)(iOplAPI.PopInt32() - sizeof(MCEXT));

	MCEXTHEADER *header = handle->header;
	TInt32 value;

	TInt idx = Lookup((MCEXTTABLE *)header->functions, funcname, &value);
	iOplAPI.Push((TInt32)idx);
}


void COpxRoot::MCLookupConst() const
	{
	// return 0 or value of constant
	TPtrC8 constname = iOplAPI.PopString8();
	MCEXT *handle = (MCEXT *)(iOplAPI.PopInt32() - sizeof(MCEXT));

	MCEXTHEADER *header = handle->header;
	TInt32 value;

	if (Lookup((MCEXTTABLE *)header->constants, constname, &value))
	  value = 0;
	iOplAPI.Push(value);
}


void COpxRoot::MCPreLoadExt() const
	{
	// return 0 or required memory
#define WORDSTOPEEK	10

	TPtrC8 filename=iOplAPI.PopString8();
	TInt filesize, required, i;
	MCEXTHEADER *header;
	TInt32 headerpeek[WORDSTOPEEK];

	header = (MCEXTHEADER *)headerpeek;

	required = 16 + sizeof(MCEXT);

	// get filesize and read the first 12 words of the file into buffer
	RFs fs;
	RFile file;
	TBuf8<4> buffer;

	fs.Connect();
	if (file.Open(fs, filename, EFileShareExclusive)) {
	        fs.Close();
		iOplAPI.Push((TInt32)-1);
		return;
	}
	file.Size(filesize);

	for (i = 0; i < WORDSTOPEEK; i++) {
		file.Read(buffer, 4);
		Mem::Copy(headerpeek+i, buffer.Ptr(), 4);
	}
	file.Close();
	fs.Close();

	// check if header is legal
	TInt32 headerfailure;
	headerfailure = is_header_legal(header, filesize);
	if (headerfailure < 0) {
		iOplAPI.Push((TInt32)headerfailure);
		return;
	}

	// update the required workspace size
	required += filesize + 4;
	required += header->workspace_size + 4;

	iOplAPI.Push((TInt32)required);
	}


void COpxRoot::MCLoadExt() const
	// return -1 or handle
	{
	TInt32 flags=iOplAPI.PopInt32();
	TPtrC8 filename=iOplAPI.PopString8();
	TInt32 opladdr=iOplAPI.PopInt32();
	MCEXT *handle;
	MCEXTHEADER *header;
	TInt *headeri;
	TInt filesize, checksum, i, oplbuffer;

	oplbuffer = (TInt32)iOplAPI.OffsetToAddrL(opladdr, 0);

	// make sure the handle is wordaligned
	handle = (MCEXT *)(oplbuffer & ~3);
	handle->oplbase = opladdr;
	handle->initialised = 0;

	// get file size and load file
	RFs fs;
	RFile file;
	TBuf8<4> buffer;

	fs.Connect();
	if (file.Open(fs, filename, EFileShareExclusive)) {
	        fs.Close();
		iOplAPI.Push((TInt32)-1);
		return;
	}
	file.Size(filesize);
	// make sure file size is a multiple of 4
	filesize = (filesize + 3) & ~3;
	// build handle
	handle->workspace = (TInt8 *)((TInt32)handle) + sizeof(MCEXT);
	handle->header = (MCEXTHEADER *)(handle->workspace + filesize + 4);
	handle->filesize = filesize;
	header = handle->header;
	headeri = (TInt *)handle->header;

	for (i = 0; i < filesize/4; i++) {
		file.Read(buffer, 4);
		Mem::Copy(headeri+i, buffer.Ptr(), 4);
	}
	file.Close();
	fs.Close();

	// StrongARM
	TInt strongarm = 0;
	if (strongarm) {
	  // flush instruction cache
	  // MCRNE  CRn=7, Rd=X, OPC_2=0, CRm=0
	  // 0x1E70f15
          asm("MCRNE p15, 0, r0, c7, c5");
	}

        { // reset list of functions
          TInt idx;
          for (idx = 0; idx < 64; idx++)    handle->functions[idx] = -1;
        }

	// update header offsets
	if (header->init.offset)
		header->init.offset += (TInt32)header;
	if (header->kill.offset)
		header->kill.offset += (TInt32)header;
	if (header->workspace_default)
  		header->workspace_default += (TInt32)header;
	if (header->functions)
		header->functions   += (TInt32)header;
	if (header->constants)
		header->constants   += (TInt32)header;
#ifdef SUPPORT_IMPORT
	if (header->importconstants)
		header->importconstants += (TInt32)header;
	if (header->importfunctions)
		header->importfunctions += (TInt32)header;
#endif
	// reset workspace
	if (header->workspace_default) {
		Mem::Copy(handle->workspace,
		          (TAny *)header->workspace_default,
 		          header->workspace_size);
	} else {
		for (i = 0; i < header->workspace_size; i++)
			handle->workspace[i] = 0;
	}
	header->workspace_default = (TInt32)handle->workspace;

	// reset functions pointer table
	if (header->functions) {
		MCEXTTABLE *t;
		TInt idx;

		t = (MCEXTTABLE *)header->functions;
		idx = 0;
		while (t->flags !=-1) {
			t->value.value += (TInt32)header;
			// save the addresses of the first 64 functions
			if (idx < 64)
			    handle->functions[idx] = t->value.value;
			idx++;
			t = skip_symbol(t);
		}
	}

	// call init
#ifdef SUPPORT_IMPORT
	if (!header->importfunctions && !header->importconstants) {
	      handle->initialised = 1;
 	      if (header->init.offset)
		    header->init.code(handle->workspace,
		                      (TInt32 *)&handle->callbackword);
	}
#else
	handle->initialised = 1;
        if (header->init.offset)
	    header->init.code(handle->workspace,
                     (TInt32 *)&handle->callbackword);
#endif

	iOplAPI.Push((TInt32)handle->workspace);
	}


void COpxRoot::MCUnloadExt() const
	{ // return 0
	MCEXT *handle = (MCEXT *)(iOplAPI.PopInt32() - sizeof(MCEXT));
	MCEXTHEADER *header = handle->header;

	if (header->kill.offset)	header->kill.code(0);
	iOplAPI.Push((TInt32)0);
	}

#ifdef SUPPORT_IMPORT
void COpxRoot::MCLink() const
	{
	// should return errorstring or "" - currently return -1 or 0
	TInt32 *mcxlist = iOplAPI.PopPtrInt32();
	TInt32 *list, failed;
        MCEXTHEADER *header;

        list = mcxlist;
        failed = 0;

        // import constants
        while ((*list) && (!failed)) {
          MCEXT *handle = (MCEXT *)(*list - sizeof(MCEXT));
	  header = handle->header;
	  if (header->importconstants)
            if (link(mcxlist, (MCEXTTABLE *)header->importconstants, header, 0))
              failed++;

	  list++;
        }
        if (failed) {
	     iOplAPI.Push((TInt32)-1);
	     return;
        }

        // import functions
        while ((*list) && (!failed)) {
          MCEXT *handle = (MCEXT *)(*list - sizeof(MCEXT));
	  header = handle->header;
	  if (header->importfunctions)
            if (link(mcxlist, (MCEXTTABLE *)header->importfunctions, header, 1))
              failed++;

	  list++;
        }
        if (failed) {
	     iOplAPI.Push((TInt32)-1);
	     return;
        }

        list = mcxlist;
        while (*list) {
          MCEXT *handle = (MCEXT *)(*list - sizeof(MCEXT));
	  header = handle->header;
	  if (header->importfunctions || header->importconstants) { // init
   	    handle->initialised = 1;
	    header->init.code(handle->workspace,
		             (TInt32 *)&handle->callbackword);
	  }
	  list++;
        }

	iOplAPI.Push((TInt32)0);
	}
#endif

void COpxRoot::MCCopyFromOpl() const
	{ // return 0
	TInt32 bytes = iOplAPI.PopInt32();
	TText8 *addr = (TText8 *)iOplAPI.PopInt32();
	TInt32 opladdr = iOplAPI.PopInt32();
	TText8 *buffer;

	buffer = iOplAPI.OffsetToAddrL(opladdr, 0);
	Mem::Copy(addr, buffer, bytes);
	iOplAPI.Push((TInt32)0);
	}


void COpxRoot::MCCopyToOpl() const
	{ // return 0
	TInt32 bytes = iOplAPI.PopInt32();
	TInt32 opladdr = iOplAPI.PopInt32();
	TText8 *addr = (TText8 *)iOplAPI.PopInt32();
	TText8 *buffer;

	buffer = iOplAPI.OffsetToAddrL(opladdr, 0);
	Mem::Copy(buffer, addr, bytes);
	iOplAPI.Push((TInt32)0);
	}

// The members of COpxRoot which are not language extension procedures
//
COpxRoot::COpxRoot(OplAPI& aOplAPI) : COpxBase(aOplAPI)
    {
     __DECLARE_NAME(_S("COpxRoot"));
    }


COpxRoot* COpxRoot::NewLC(OplAPI& aOplAPI)
    {
    COpxRoot* This=new(ELeave) COpxRoot(aOplAPI);
    CleanupStack::PushL(This);
    This->ConstructL();
    return This;
    }


void COpxRoot::ConstructL()
    {
	// Do whatever is required to construct any component members you have in COpxRoot
	// ...

    }

COpxRoot::~COpxRoot()
    {
    Dll::FreeTls();		// Required so that Tls is set to zero on unloading the OPX in UNLOADM
    }

// COpxBase implementation
//
void COpxRoot::RunL(TInt aProcNum)
// Run a language extension procedure
	{
	switch (aProcNum)
		{
	case EMCPRELOADEXT:
		MCPreLoadExt();
		break;
	case EMCLOADEXT:
		MCLoadExt();
		break;
	case EMCUNLOADEXT:
		MCUnloadExt();
		break;
	case EMCLOOKUP:
		MCLookup();
		break;
	case EMCLOOKUPCONST:
		MCLookupConst();
		break;
	case EMCCALL:
		MCCall();
		break;
	case EMCCALLF:
		MCCallF();
		break;
	case EMCCALLADDR:
		MCCallAddr();
		break;
	case EMCPEEKL:
		MCPeekL();
		break;
	case EMCPEEKW:
		MCPeekW();
		break;
	case EMCPEEKB:
		MCPeekB();
		break;
	case EMCPOKEL:
		MCPokeL();
		break;
	case EMCPOKEW:
		MCPokeW();
		break;
	case EMCPOKEB:
		MCPokeB();
		break;
	case EMCPOKES:
		MCPokeS();
		break;
	case EMCCOPYTOOPL:
		MCCopyToOpl();
		break;
	case EMCCOPYFROMOPL:
		MCCopyFromOpl();
		break;
	case EMCOPLADDR:
		MCOplAddr();
		break;
#ifdef SUPPORT_IMPORT
        case EMCLINK:
		MCLink();
		break;
#endif
	default:
		User::Leave(KOplErrOpxProcNotFound);
		}
	}

TBool COpxRoot::CheckVersion(TInt aVersion)
// To check whether the opx is a compatible version
// *** Change as required ***
	{
	if ((aVersion & 0xff00)>(KOpxVersion & 0xff00))	// major version must be <= OPX's version
		return EFalse; // bad version
	else
		return ETrue; // ok
	}


EXPORT_C COpxBase* NewOpxL(OplAPI& aOplAPI)
// Creates a COpxBase instance as required by the OPL runtime
// This object is to be stored in the OPX's TLS as shown below
	{
	COpxRoot* tls=((COpxRoot*)Dll::Tls());
	if (tls==NULL)		// tls is NULL on loading an OPX DLL (also after unloading it)
		{
        tls=COpxRoot::NewLC(aOplAPI);
	    User::LeaveIfError(Dll::SetTls(tls));
		CleanupStack::Pop();	// tls
        }
    return (COpxBase *)tls;
	}


EXPORT_C TUint Version()
// ER5 and later: export the OPX version number.
	{
	return KOpxVersion;
	}


GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
//
// DLL entry point
//
	{
	return(KErrNone);
	}

//------------------------------------------------------------------------


void _decrypt(TInt32 *data, TInt32 size, TInt32 key) {
}


TInt32 is_header_legal(MCEXTHEADER *header, TInt32 filesize)
	{

	if (header->workspace_size < 4 || header->workspace_size > 2000000) {
		return -1;
	}
	if (header->terminator != -1) {
		return -2;
	}
	if ((header->init.offset < 32 && header->init.offset != 0) ||
	     header->init.offset > filesize) {
		return -3;
	}
	if ((header->kill.offset < 32 && header->kill.offset != 0) ||
             header->kill.offset > filesize) {
		return -4;
	}
	if (((header->workspace_default < 32) && (header->workspace_default != 0)) ||
	    (header->workspace_default > filesize) ||
	    (header->workspace_default &3)) {
		return -5;
	}
	if (((header->functions < 32) && (header->functions != 0)) ||
 	     (header->functions > filesize) ||
	     (header->functions &3)) {
		return -6;
	}
	if (((header->constants < 32) && (header->constants != 0)) ||
	     (header->constants > filesize) ||
	     (header->constants &3)) {
		return -7;
	}

	return 0;	// OK
}



TInt Lookup(MCEXTTABLE *t, TPtrC8 name, TInt32 *value)
// return index or -1 if not found
	{
	TInt namelen = name.Length();
	const TText8 *namep = name.Ptr();
	TInt idx;

	if (!t)		return -1;
	idx = 0;
	while (t->flags != -1) {
		// skip value and flags
		if (_strcmp(t->name, namep, namelen) == 0) {
		    *value = t->value.value;
	            return idx;
		}
		t = skip_symbol(t);
		idx++;
	}

	return -1;
	}


TInt LookupN(MCEXTTABLE *t, TInt32 idx, TInt32 *value)
	// return 0 or the value
	{

	if (!t)		return 0;
	while ((t->flags != -1) && (idx)) {
	  t = skip_symbol(t);
	  idx--;
	}

	if (t->flags == -1)   return 1;
	*value = t->value.value;
	return 0;
	}

#ifdef SUPPORT_IMPORT
TInt link(TInt32 *extlist, MCEXTTABLE *import, MCEXTHEADER *header, TInt b) {
// return 0 for OK or -1 for error
  TInt32 *list, success;

   success = 0;
   // for each symbol in the import list
   while ((import->flags != -1) && (!success)) {
       list = extlist;
       // for each possible extension
       while (*list) {
         MCEXT *linkext = (MCEXT *)(*list - sizeof(MCEXT));
         MCEXTHEADER *extheader = linkext->header;
         // attempt to link
         if (extheader != header) {
           if (b) {
             if (extheader->importfunctions)
                 if (!link_symbol((MCEXTTABLE *)extheader->importfunctions,
                                 import, header, b))       success = 1;
           } else {
             if (extheader->importconstants)
                 if (!link_symbol((MCEXTTABLE *)extheader->importconstants,
                                 import, header, b))       success = 1;
           }
         }
         list++;
       }
   }
   if (success)   return 0;
   return 1;
}


TInt link_symbol(MCEXTTABLE *import, MCEXTTABLE *symbol, MCEXTHEADER *header, TInt b) {
  // return 0 for OK or -1 for error
  TText8 *name, *t;
  TInt32 flags, value, i;

  flags = symbol->flags;
  value = symbol->value.value;
  name = symbol->name;

  while (import->flags != -1) {
    if (_strcmp(name, import->name, -1) == 0) {
      // match!!
      // HBP
      return 0;
    }
    import = skip_symbol(import);
  }
  return -1;
}

#endif

TInt _strcmp(const TText8 *t1, const TText8 *t2, const TInt len2) {
// return 0 for equal or <>0 for not equal
  TInt i;
  // compare string at p with string to match
  if (len2 == -1) {
    for (i = 0; (t1[i]) && (t2[i]); i++)
	if (t1[i] != t2[i])	return 1;
    return 0;

  } else {
    for (i = 0; (t1[i]) && (i < len2); i++) {
	if (t1[i] != t2[i])	return 1;
	if ((t1[i+1] == '\0') && (i == len2-1))   return 0;
    }
  }

  return 1;
}


MCEXTTABLE *skip_symbol(MCEXTTABLE *symbol) {
   // skip name
   TText8 *p = symbol->name;
   while (*p)    p++;
   // skip terminating 0
   p++;
   // word align
   while (((TInt32)p) &3)  p++;

   return (MCEXTTABLE *)p;
}
