// ASMX.H
//
//
#if !defined(__MC_H__)
#define __MC_H__

#if !defined(__E32BASE_H__)
#include <e32base.h>
#endif
#if !defined(__OPLAPI_H__)
#include <oplapi.h>
#endif
#if !defined(__OPX_H__)
#include <opx.h>
#endif

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// The OPX's UID as used in the DECLARE OPX statement for your OPX
// NB. This UID must be reserved by contacting Symbian Ltd
//
const TInt KUidOpxMC=0x1000a5cd;


// The OPX's version number as used in the DECLARE OPX statement for your OPX
const TInt KOpxVersion=0x100;

// the struture holding info about a loaded extension
typedef struct MCEXT MCEXT;
typedef struct MCEXTHEADER MCEXTHEADER;
typedef struct MCEXTTABLE MCEXTTABLE;


struct MCEXT {
  MCEXTHEADER *header;
  TInt32 callbackword;
  TInt32 initialised;
  TInt8 *workspace;
  TInt32 filesize;
  TInt32 oplbase;
  TInt32 functions[64];
};

struct MCEXTTABLE {
  TInt32 flags;
  union {
    TInt32 value;
    TInt32 (* code)(TInt32, TInt32, TInt32, TInt32, TInt32 );
  } value;
  TText8 name[256];
};


struct MCEXTHEADER {
  TInt32 id;
  TInt32 flags;
  TInt32 workspace_size;
  TInt32 workspace_default;
  union {
    TInt32 (* code)(TInt8 *, TInt32 *);
    TInt32 offset;
  } init;
  union {
    TInt32 (* code)(TInt8 *);
    TInt32 offset;
  } kill;
  TInt32 functions;
  TInt32 constants;
  TInt32 terminator;
};


// The root class of the OPX - add any members and/or component classes
// that your particular OPX requires to this class
//
// An instance of this class is created immediately after the OPX is loaded (using LOADM)
// and the object pointer must be stored in the Tls (Thread local static).
// See the 1st EXPORT function NewOpxL(), which does this for you.
// The Tls contains the single global (32-bit) value available to any EPOC32 DLL.
// No other non-const static data may be used.
//
class COpxRoot : public COpxBase
    {
public:
    static COpxRoot* NewLC(OplAPI& aOplAPI);
	virtual void RunL(TInt aProcNum);
	virtual TInt CheckVersion(TInt aVersion);
	// Add any other members specific to your OPX
	// ...
private:
	// the language extension procedures
	enum TExtensions
		{
		EMCPRELOADEXT=1,
		EMCLOADEXT,
		EMCUNLOADEXT,
		EMCLOOKUP,
		EMCLOOKUPCONST,
		EMCCALL,
		EMCCALLF,
		EMCCALLADDR,
		EMCPEEKL,
		EMCPEEKW,
		EMCPEEKB,
		EMCPOKEL,
		EMCPOKEW,
		EMCPOKEB,
		EMCPOKES,
		EMCCOPYTOOPL,
		EMCCOPYFROMOPL,
		EMCOPLADDR
		};
	void MCPreLoadExt() const;
	void MCLoadExt() const;
	void MCUnloadExt() const;
	void MCLookup() const;
	void MCLookupConst() const;
	void MCCall() const;
	void MCCallF() const;
	void MCCallAddr() const;
	void MCPeekL() const;
	void MCPeekW() const;
	void MCPeekB() const;
	void MCPokeL() const;
	void MCPokeW() const;
	void MCPokeB() const;
	void MCPokeS() const;
	void MCCopyToOpl() const;
	void MCCopyFromOpl() const;
	void MCOplAddr() const;
private:
	void ConstructL();
	COpxRoot(OplAPI& aOplAPI);
	~COpxRoot();
    };



// OpxRoot() accesses the Tls which contains the COpxRoot*, created and stored there by NewOpxL()
//
// OpxRoot() is only required if you need to access your COpxRoot object
// in a deeply nested function which hasn't been passed this pointer directly
// Note that accessing the Tls in this way is relatively slow (about 4 microsecs) so
// it is often more efficient to pass the COpxRoot* down the chain of functions.
inline COpxRoot* OpxRoot() { return((COpxRoot *)Dll::Tls()); }


#endif __ASMX_H__
