// WGFX.CPP
//



#include <oplerr.h>
#include <libc/stdio.h>
#include <fbs.h>
#include <gdi.h>
#include <opx.h>
#include <w32std.h>
#include <coemain.h>

#include "wgfx.h"

void DeleteBitmap(OPLBITMAP *bitmap);
TInt ConvertIDtoIndex(TInt);
TInt ConvertIndextoID(TInt);

//////////////////////////////////////////////////////////////////////////
// CWGFXOpx class

CWGFXOpx::CWGFXOpx()
      {
      __DECLARE_NAME(_S("CWGFXOpx"));
      TInt i;

      for (i = 0; i < 64; i++)  bitmaps[i].bitmap = NULL;
      }


CWGFXOpx::~CWGFXOpx()
      {
      TInt i;

      for (i = 0; i < 64; i++)
        if (bitmaps[i].bitmap)      DeleteBitmap(bitmaps+i);
      }


void CWGFXOpx::WGFXCreate(OplAPI &aOplAPI)
	{
	TInt32 *obpl = aOplAPI.PopPtrInt32();
	TInt32 bpp = aOplAPI.PopInt32();
	TInt32 sizey = aOplAPI.PopInt32();
	TInt32 sizex = aOplAPI.PopInt32();

	TInt i, bi;
	bi = -1;
	for (i = 0; (i < 64) && (bi == -1); i++) {
	  if ((bitmaps[i].bitmap == NULL) &&
	      (bitmaps[i].locked & SLOT_LOCKED) == 0) {
	    bitmaps[i].locked |= SLOT_LOCKED;
            bi = i;
	  }
	}

        if (bi == -1) {
               bitmaps[bi].locked &= !SLOT_LOCKED;
               aOplAPI.Push((TInt32)0);
               return;
        }

        if (bpp != 2 && bpp != 4 && bpp != 8) {
              bitmaps[bi].locked &= !SLOT_LOCKED;
              aOplAPI.Push((TInt32)0);
              return;
        }

        // create bitmap
        CFbsBitmap *ibitmap = new (ELeave) CFbsBitmap();
        TInt ec, bpl;
        if (bpp == 2) {
          ec = ibitmap->Create(TSize(sizex,sizey), EGray4);
          if (!ec) {
            bpl = ibitmap->ScanLineLength(sizex, EGray4);
            aOplAPI.PutLong(obpl,bpl);
          }
        } else if (bpp == 4) {
          ec = ibitmap->Create(TSize(sizex,sizey), EGray16);
          if (!ec) {
            bpl = ibitmap->ScanLineLength(sizex, EGray16);
            aOplAPI.PutLong(obpl,bpl);
          }
        } else if (bpp == 8) {
          ec = ibitmap->Create(TSize(sizex,sizey), EGray256);
          if (!ec) {
            bpl = ibitmap->ScanLineLength(sizex, EGray256);
            aOplAPI.PutLong(obpl,bpl);
          }
        }

        if (ec) {
               bitmaps[bi].locked &= !SLOT_LOCKED;
     	       aOplAPI.Push((TInt32)0);
 	       return;
        }

        // create graphics context
        CWsScreenDevice *sd = aOplAPI.ScreenDevice();
        CWindowGc *igc;
        if (sd->CreateContext(igc)) {
               bitmaps[bi].locked &= !SLOT_LOCKED;
               ibitmap->Reset();
               delete ibitmap;
               aOplAPI.Push((TInt32)0);
               return;
        }

        TBitmapUtil *ibutil = new (ELeave) TBitmapUtil(ibitmap);

        bitmaps[bi].bitmap = ibitmap;
        bitmaps[bi].butil = ibutil;
        bitmaps[bi].gc = igc;

        bitmaps[bi].sizex = sizex;
        bitmaps[bi].sizey = sizey;
        bitmaps[bi].bpp = bpp;
        bitmaps[bi].totalsize = sizey*bpl;

        bitmaps[bi].locked &= !SLOT_LOCKED;

	aOplAPI.Push((TInt32)ConvertIndextoID(bi));
	}


void CWGFXOpx::WGFXInfo(OplAPI &aOplAPI)
	{
	TInt bi = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }

        if (!bitmaps[bi].bitmap) {
              aOplAPI.Push((TInt32)0);
              return;
        }

	aOplAPI.Push((TInt32)bitmaps[bi].bitmap->DataAddress());
	}


void CWGFXOpx::WGFXCopy(OplAPI &aOplAPI)
	{
	const TAny *buf = (TAny *)aOplAPI.PopInt32();
	TInt bi = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }

        if (!bitmaps[bi].bitmap) {
              aOplAPI.Push((TInt32)0);
              return;
        }

        if (!(bitmaps[bi].locked & BITMAP_LOCKED)) {
              bitmaps[bi].locked |= BITMAP_LOCKED;
              bitmaps[bi].butil->Begin(TPoint(0,0));
        }

        TAny *bibuf = bitmaps[bi].bitmap->DataAddress();
        Mem::Move(bibuf, buf, bitmaps[bi].totalsize);

        if (bitmaps[bi].locked & BITMAP_LOCKED) {
              bitmaps[bi].locked &= ~BITMAP_LOCKED;
              bitmaps[bi].butil->End();
        }

	aOplAPI.Push((TInt32)0);
	}


void CWGFXOpx::WGFXCopyPlot(OplAPI &aOplAPI)
	{
        TInt32 y = aOplAPI.PopInt32();
        TInt32 x = aOplAPI.PopInt32();
	const TAny *buf = (TAny *)aOplAPI.PopInt32();
	TInt bi = (TInt)aOplAPI.PopInt32();
	TInt wid = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }

        if (!bitmaps[bi].bitmap) {
              aOplAPI.Push((TInt32)0);
              return;
        }

        if (!(bitmaps[bi].locked & BITMAP_LOCKED)) {
              bitmaps[bi].locked |= BITMAP_LOCKED;
              bitmaps[bi].butil->Begin(TPoint(0,0));
        }

        TAny *bibuf = bitmaps[bi].bitmap->DataAddress();
        Mem::Move(bibuf, buf, bitmaps[bi].totalsize);

        if (bitmaps[bi].locked & BITMAP_LOCKED) {
              bitmaps[bi].locked &= ~BITMAP_LOCKED;
              bitmaps[bi].butil->End();
        }

        RBackedUpWindow iwin = aOplAPI.WindowFromIdL(wid);

        bitmaps[bi].gc->Activate(iwin);
        bitmaps[bi].gc->BitBlt(TPoint(x, y), bitmaps[bi].bitmap);
        bitmaps[bi].gc->Deactivate();

	aOplAPI.Push((TInt32)0);
	}


void CWGFXOpx::WGFXLock(OplAPI &aOplAPI)
	{
	TInt bi = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }

        if (!bitmaps[bi].bitmap) {
              aOplAPI.Push((TInt32)0);
              return;
        }

        if (!(bitmaps[bi].locked & BITMAP_LOCKED)) {
              bitmaps[bi].locked |= BITMAP_LOCKED;
              bitmaps[bi].butil->Begin(TPoint(0,0));
        }

	aOplAPI.Push((TInt32)1);
	}


void CWGFXOpx::WGFXUnlock(OplAPI &aOplAPI)
	{
	TInt bi = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }

        if (!bitmaps[bi].bitmap) {
              aOplAPI.Push((TInt32)0);
              return;
        }

        if (bitmaps[bi].locked & BITMAP_LOCKED) {
              bitmaps[bi].locked &= ~BITMAP_LOCKED;
              bitmaps[bi].butil->End();
        }

	aOplAPI.Push((TInt32)1);
	}


void CWGFXOpx::WGFXDestroy(OplAPI &aOplAPI)
	{
	TInt bi = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }
        DeleteBitmap(bitmaps+bi);

	aOplAPI.Push((TInt32)1);
	}


void CWGFXOpx::WGFXPlot(OplAPI &aOplAPI)
	{
        TInt32 y = aOplAPI.PopInt32();
        TInt32 x = aOplAPI.PopInt32();
	TInt bi = (TInt)aOplAPI.PopInt32();
	TInt wid = (TInt)aOplAPI.PopInt32();

        bi = ConvertIDtoIndex(bi);
        if (bi < 0) {
	      aOplAPI.Push((TInt32)0);
	      return;
        }
        RBackedUpWindow iwin = aOplAPI.WindowFromIdL(wid);

        bitmaps[bi].gc->Activate(iwin);
        bitmaps[bi].gc->BitBlt(TPoint(x, y), bitmaps[bi].bitmap);
        bitmaps[bi].gc->Deactivate();

	aOplAPI.Push((TInt32)1);
	}


// The members of COpxRoot which are not language extension procedures
//
COpxRoot::COpxRoot(OplAPI& aOplAPI) : COpxBase(aOplAPI)
    {
     __DECLARE_NAME(_S("COpxRoot"));
    }


COpxRoot* COpxRoot::NewLC(OplAPI& aOplAPI)
    {
    COpxRoot* This=new(ELeave) COpxRoot(aOplAPI);
    CleanupStack::PushL(This);
    This->ConstructL();
    return This;
    }


void COpxRoot::ConstructL()
    {
      wgfxhandle = new (ELeave) CWGFXOpx;
	// Do whatever is required to construct any component members you have in COpxRoot
	// ...

    }

COpxRoot::~COpxRoot()
    {
      delete wgfxhandle;
      Dll::FreeTls();		// Required so that Tls is set to zero on unloading the OPX in UNLOADM
    }

// COpxBase implementation
//
void COpxRoot::RunL(TInt aProcNum)
// Run a language extension procedure
	{
	switch (aProcNum)
		{
	case EWGFXCREATE:
		wgfxhandle->WGFXCreate(iOplAPI);
		break;
        case EWGFXLOCK:
                wgfxhandle->WGFXLock(iOplAPI);
                break;
        case EWGFXINFO:
                wgfxhandle->WGFXInfo(iOplAPI);
                break;
        case EWGFXCOPY:
                wgfxhandle->WGFXCopy(iOplAPI);
                break;
        case EWGFXPLOT:
                wgfxhandle->WGFXPlot(iOplAPI);
                break;
	case EWGFXUNLOCK:
		wgfxhandle->WGFXUnlock(iOplAPI);
		break;
        case EWGFXDESTROY:
                wgfxhandle->WGFXDestroy(iOplAPI);
                break;
	default:
		User::Leave(KOplErrOpxProcNotFound);
		}
	}

TBool COpxRoot::CheckVersion(TInt aVersion)
// To check whether the opx is a compatible version
// *** Change as required ***
	{
	if ((aVersion & 0xff00)>(KOpxVersion & 0xff00))	// major version must be <= OPX's version
		return EFalse; // bad version
	else
		return ETrue; // ok
	}


EXPORT_C COpxBase* NewOpxL(OplAPI& aOplAPI)
// Creates a COpxBase instance as required by the OPL runtime
// This object is to be stored in the OPX's TLS as shown below
	{
	COpxRoot* tls=((COpxRoot*)Dll::Tls());
	if (tls==NULL)		// tls is NULL on loading an OPX DLL (also after unloading it)
		{
        tls=COpxRoot::NewLC(aOplAPI);
	    User::LeaveIfError(Dll::SetTls(tls));
		CleanupStack::Pop();	// tls
        }
    return (COpxBase *)tls;
	}


EXPORT_C TUint Version()
// ER5 and later: export the OPX version number.
	{
	return KOpxVersion;
	}


GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
//
// DLL entry point
//
	{
	return(KErrNone);
	}


// ------------------------------------------------------------

void DeleteBitmap(OPLBITMAP *bitmap) {

  CFbsBitmap *ibitmap = bitmap->bitmap;
  TBitmapUtil *ibutil = bitmap->butil;
  CWindowGc *igc      = bitmap->gc;

  bitmap->bitmap   = NULL;

  if (bitmap->locked & BITMAP_LOCKED)  ibutil->End();
  delete ibutil;
  delete igc;
  ibitmap->Reset();   // delete bitmap
  delete ibitmap;
  bitmap->locked = 0;
}


TInt ConvertIndextoID(TInt i) {
  return 17+128*i;
}


TInt ConvertIDtoIndex(TInt i) {
  i = (i-17)/128;
  if (i < 0 || i >= 64)   return -1;
  return i;
}

