// WGFX.H
//
//
#if !defined(__WGFX_H__)
#define __WGFX_H__

#if !defined(__E32BASE_H__)
#include <e32base.h>
#endif
#if !defined(__OPLAPI_H__)
#include <oplapi.h>
#endif
#if !defined(__OPX_H__)
#include <opx.h>
#endif


#define BITMAP_LOCKED         1
#define SLOT_LOCKED           2

typedef struct OPLBITMAP {
  CFbsBitmap *bitmap;
  CWindowGc *gc;
  TBitmapUtil *butil;
  TInt32 locked, sizex, sizey, totalsize, bpp;
} OPLBITMAP;

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// The OPX's UID as used in the DECLARE OPX statement for your OPX
// NB. This UID must be reserved by contacting Symbian Ltd
//
const TInt KUidOpxWGFX=0x1000a5ce;


// The OPX's version number as used in the DECLARE OPX statement for your OPX
const TInt KOpxVersion=0x100;


// The root class of the OPX - add any members and/or component classes
// that your particular OPX requires to this class
//
// An instance of this class is created immediately after the OPX is loaded (using LOADM)
// and the object pointer must be stored in the Tls (Thread local static).
// See the 1st EXPORT function NewOpxL(), which does this for you.
// The Tls contains the single global (32-bit) value available to any EPOC32 DLL.
// No other non-const static data may be used.
//

class CWGFXOpx : public CBase {
  public:
        CWGFXOpx();
        ~CWGFXOpx();
        void WGFXCreate(OplAPI& aOplAPI);
	void WGFXLock(OplAPI& aOplAPI);
	void WGFXInfo(OplAPI& aOplAPI);
	void WGFXCopy(OplAPI& aOplAPI);
	void WGFXCopyPlot(OplAPI& aOplAPI);
	void WGFXPlot(OplAPI& aOplAPI);
	void WGFXUnlock(OplAPI& aOplAPI);
	void WGFXDestroy(OplAPI& aOplAPI);

        OPLBITMAP bitmaps[64];
};


class COpxRoot : public COpxBase
    {
public:
    static COpxRoot* NewLC(OplAPI& aOplAPI);
	virtual void RunL(TInt aProcNum);
	virtual TInt CheckVersion(TInt aVersion);
        CWGFXOpx *wgfxhandle;
private:
	// the language extension procedures
	enum TExtensions
		{
		EWGFXCREATE=1,
		EWGFXLOCK,
		EWGFXINFO,
		EWGFXCOPY,
		EWGFXCOPYPLOT,
		EWGFXPLOT,
		EWGFXUNLOCK,
		EWGFXDESTROY,
		};
private:
	void ConstructL();
	COpxRoot(OplAPI& aOplAPI);
	~COpxRoot();
    };



// OpxRoot() accesses the Tls which contains the COpxRoot*, created and stored there by NewOpxL()
//
// OpxRoot() is only required if you need to access your COpxRoot object
// in a deeply nested function which hasn't been passed this pointer directly
// Note that accessing the Tls in this way is relatively slow (about 4 microsecs) so
// it is often more efficient to pass the COpxRoot* down the chain of functions.
inline COpxRoot* OpxRoot() { return((COpxRoot *)Dll::Tls()); }


#endif __WGFX_H__
